----------------------------
    -- Require
----------------------------
cpml = require "libs/cpml"
l3d  = require "libs/love3d"
mesh = require "libs/mesh"

----------------------------
    -- Locals
----------------------------
local w, h = love.graphics.getDimensions()
local intersect = false
local elapsed   = 0

----------------------------
    -- Camera
----------------------------
local camera = {
    -- General data
    position    = cpml.vec3( 0.5, -1 , -3  ),
    rotation    = cpml.vec3(  0 , -1 ,-0.2 ),
    orientation = cpml.quat(  0 ,  0 ,  0  , 1  ),
    
    -- Orientation axis
    up = cpml.vec3(0, 0, 1),
    
    -- Field of view
    field_of_view = 70,
    
    -- Shader parameters
    view        = cpml.mat4(),
    projection  = cpml.mat4(),
}

-- Send camera data to shader
function camera:send_to_shader()
    -- Process data
    camera.view = camera.view:identity()
        :translate(camera.position)
        :look_at(cpml.vec3(0, 0, 0), camera.rotation, camera.up)
    
    camera.projection = camera.projection:identity()
        :perspective(camera.field_of_view, w/h, 0.001, 1000)
    
    -- Send data
    love.graphics.getShader():send("u_view"      , camera.view:to_vec4s()      )
    love.graphics.getShader():send("u_projection", camera.projection:to_vec4s())
end

----------------------------
    -- Main
----------------------------
function love.load()
    -- Patch Love2D
    l3d.import(true)

    -- Initialize geometry
    AABB = {
        min = cpml.vec3(0, 0, 0), 
        max = cpml.vec3(1, 1, 1)
    }
    
    triangle = {
        cpml.vec3(0, 0.5, -0.5),    -- Vertex 1
        cpml.vec3(1, 0.5, -0.5),    -- Vertex 2
        cpml.vec3(0, 0.5,  0.5)     -- Vertex 3
    }
    
    -- Load 3D shader
    shader = love.graphics.newShader("shader.glsl")
    
    -- Set mesh default texture
    texture = love.graphics.newImage("texture.png")
    mesh.texture_default(texture)
    
    -- Generate rendering meshes
    local size = AABB.max - AABB.min
    
    AABB_mesh = mesh.cuboid_new(
        cpml.vec3(),
        cpml.vec3(size.x, size.y, size.z),
        {{x = 0, y = 0, w = 256, h = 256}}
    )
    
    triangle_mesh = tri_new(triangle)

    love.graphics.setWireframe(false)
end

function love.update(dt)
    elapsed = elapsed + dt
    
    -- Move AABB
    local new_position = math.sin(elapsed)*2
    
    -- Note: add move and move_to functions to the mesh module
    AABB.min = cpml.vec3(new_position, 0, 0)
    AABB.max = cpml.vec3(1 + new_position, 1, 1)
    AABB_mesh.position = cpml.vec3(new_position, 0, 0)
    
    local a, b, c = unpack(triangle)
    intersect = aabb_triangle(a, b, c, AABB)
end

function love.draw()
	love.graphics.push("all")

	-- Depth buffer
	love.graphics.clearDepth()
	love.graphics.setDepthTest("less")

    love.graphics.setShader(shader)
    camera:send_to_shader()
	
	-- Draw geometry
    mesh.draw(AABB_mesh)
    triangle_draw()
    
	-- Reset depth test
	love.graphics.setDepthTest()
	love.graphics.setShader()
    
    -- Print the intersection test's value
    love.graphics.print("Intersection: " .. tostring(intersect))
    
    love.graphics.pop()
end

----------------------------
    -- Triangle Mesh
----------------------------
function tri_new(coords)
    -- Our quad object
    local tri = {}

    -- Default minimum vertex position to 0
    local min = cpml.vec3()

    -- Vertex coordinates
	local vertices = {
        {coords[1].x, coords[1].y, coords[1].z},
        {coords[2].x, coords[2].y, coords[2].z},
        {coords[3].x, coords[3].y, coords[3].z},
	}

    -- Create the mesh
	local layout = {{ "VertexPosition", "float", 3 }}
	tri = love.graphics.newMesh(layout, vertices, "triangles", "static")

    return tri
end

function triangle_draw()
    local projection = cpml.mat4():translate(cpml.vec3())
    love.graphics.getShader():send("u_model", projection:to_vec4s())

    -- draw the mesh
    love.graphics.draw(triangle_mesh)
end

----------------------------
    -- Collision Test
----------------------------
-- From https://gist.github.com/yomotsu/d845f21e2e1eb49f647f
function aabb_triangle(a, b, c, aabb)
    local p0, p1, p2, r

    -- Compute box center and extents of AABoundingBox (if not already given in that format)
    local center  = (aabb.max + aabb.min)*0.5
    local extents =  aabb.max - center

    -- Translate triangle as conceptually moving AABB to origin
    local v0 = a - center
    local v1 = b - center
    local v2 = c - center

    -- Compute edge vectors for triangle
    local f0 = v1 - v0
    local f1 = v2 - v1
    local f2 = v0 - v2

    -- Test axes a00..a22 (category 3)
    local a00 = cpml.vec3(   0   , -f0.z , f0.y  )
    local a01 = cpml.vec3(   0   , -f1.z , f1.y  )
    local a02 = cpml.vec3(   0   , -f2.z , f2.y  )
    local a10 = cpml.vec3( f0.z  ,   0   , -f0.x )
    local a11 = cpml.vec3( f1.z  ,   0   , -f1.x )
    local a12 = cpml.vec3( f2.z  ,   0   , -f2.x )
    local a20 = cpml.vec3( -f0.y , f0.x  ,   0   )
    local a21 = cpml.vec3( -f1.y , f1.x  ,   0   )
    local a22 = cpml.vec3( -f2.y , f2.x  ,   0   )

    -- Test axis a00
    p0 = v0:dot( a00 )
    p1 = v1:dot( a00 )
    p2 = v2:dot( a00 )
    r = extents.y * math.abs( f0.z ) + extents.z * math.abs( f0.y )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a01
    p0 = v0:dot( a01 )
    p1 = v1:dot( a01 )
    p2 = v2:dot( a01 )
    r = extents.y * math.abs( f1.z ) + extents.z * math.abs( f1.y )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a02
    p0 = v0:dot( a02 )
    p1 = v1:dot( a02 )
    p2 = v2:dot( a02 )
    r = extents.y * math.abs( f2.z ) + extents.z * math.abs( f2.y )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a10
    p0 = v0:dot( a10 )
    p1 = v1:dot( a10 )
    p2 = v2:dot( a10 )
    r = extents.x * math.abs( f0.z ) + extents.z * math.abs( f0.x )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a11
    p0 = v0:dot( a11 )
    p1 = v1:dot( a11 )
    p2 = v2:dot( a11 )
    r = extents.x * math.abs( f1.z ) + extents.z * math.abs( f1.x )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a12
    p0 = v0:dot( a12 )
    p1 = v1:dot( a12 )
    p2 = v2:dot( a12 )
    r = extents.x * math.abs( f2.z ) + extents.z * math.abs( f2.x )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a20
    p0 = v0:dot( a20 )
    p1 = v1:dot( a20 )
    p2 = v2:dot( a20 )
    r = extents.x * math.abs( f0.y ) + extents.y * math.abs( f0.x )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a21
    p0 = v0:dot( a21 )
    p1 = v1:dot( a21 )
    p2 = v2:dot( a21 )
    r = extents.x * math.abs( f1.y ) + extents.y * math.abs( f1.x )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test axis a22
    p0 = v0:dot( a22 )
    p1 = v1:dot( a22 )
    p2 = v2:dot( a22 )
    r = extents.x * math.abs( f2.y ) + extents.y * math.abs( f2.x )
    if ( math.max( -math.max( p0, p1, p2 ), math.min( p0, p1, p2 ) ) > r ) then
        return false -- Axis is a separating axis
    end

    -- Test the three axes corresponding to the face normals of AABB (category 1).
    if math.max(v0.x, v1.x, v2.x) < -extents.x or math.min(v0.x, v1.x, v2.x) > extents.x then
        return false
    end
    if math.max(v0.y, v1.y, v2.y) < -extents.y or math.min(v0.y, v1.y, v2.y) > extents.y then
        return false
    end
    if math.max(v0.z, v1.z, v2.z) < -extents.z or math.min(v0.z, v1.z, v2.z) > extents.z then
        return false
    end

    -- Test separating axis corresponding to triangle face normal (category 2)
    local triangle_normal   = f1:clone():cross(f0):normalize()
    local triangle_constant = triangle_normal:dot(a)

    local r = extents.x * math.abs( triangle_normal.x ) +
              extents.y * math.abs( triangle_normal.y ) +
              extents.z * math.abs( triangle_normal.z )
              
    -- triangle_normal
    local s = triangle_normal:dot( center ) - triangle_constant

    return math.abs( s ) <= r
end